#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The module calculates scores based on information theory.

Created on Thu Mar 17 14:53:28 2022

@author: jsteiglechner
"""

import torch


def contingency_table_fn(
    probs: torch.Tensor,
    probs_prime: torch.Tensor,
) -> torch.Tensor:
    r"""
    Calculate the contingency table of the pair probs, probs_prime.

    Notes
    -----
        The contingency table is a K x K^\prime matrix, whose kk^\prime-th
        element is the number of points in the intersection of clusters C_k of
        C and C_{k^\prime}^\prime of C^\prime.

        .. math:

            n_{k k^\prime} = \left| C_k \cap C_{k^\prime}^\prime \right|

    Parameters
    ----------
    probs : torch.Tensor
        A segmentation or clustering C with K segments with shape N x K x SP
    probs_prime : torch.Tensor
        A segmentation or clustering C^\prime with K^\prime segments with shape
        N x K^\prime x SP.

    Returns
    -------
    contingency_table : torch.Tensor
        The contingency table is a N x K x K^\prime tensor.

    """
    dims = tuple(torch.arange(start=1, end=probs.dim() - 1))

    contingency_table = torch.empty(
        (
            probs.shape[0],
            probs.shape[1],
            probs_prime.shape[1],
        ),
        dtype=torch.float32,
        device=probs.device,
        requires_grad=probs.requires_grad,
    )

    for i_label, i_label_prime in torch.cartesian_prod(
        torch.arange(probs.shape[1], device=probs.device),
        torch.arange(probs_prime.shape[1], device=probs.device),
    ):
        contingency_table[:, i_label, i_label_prime] = torch.sum(
            probs[:, i_label] * probs_prime[:, i_label_prime],
            dim=dims,
        )

    return contingency_table


def marginal_entropy_fn(probs: torch.Tensor) -> torch.Tensor:
    r"""
    Calculate the entropy of a random variable.

    Notes
    -----
        We call H(C) the entropy associated with clustering C with
        probabilities P(k) of the outcome being in cluster C_k.

        .. math:
            H(C) = - \sum_{k=1} P(k) \log P(k)
            P(k) = \frac{n_k}{n}

    Parameters
    ----------
    probs : torch.Tensor
        A segmentation or clustering C with K segments with shape N x K x SP.

    Returns
    -------
    entropy : torch.Tensor
        Entropy of C with shape N.

    """
    dims = tuple(torch.arange(start=2, end=probs.dim()))
    num_elements = torch.prod(torch.tensor(probs.shape[2:]))

    probability_distribution = torch.sum(probs, dim=dims) / num_elements
    entropy = -torch.sum(
        probability_distribution * torch.log(probability_distribution),
        dim=(1),
    )

    return entropy


def joint_entropy_fn(
    probs: torch.Tensor,
    probs_prime: torch.Tensor,
    smooth: float = 10e-6,
) -> torch.Tensor:
    r"""
    Calculate the joint entropy of the random variables.

    Notes
    -----
        Represent the probability that a point belongs to C_k in clustering C
        and to C_{k^\prime}^\prime in C^\prime. H (C, C^\prime) is the entropy
        of P (k, k^\prime)

        .. math:

            H (C, C^\prime) & = - \sum_{k=1}^K \sum_{k^\prime = 1}^{K^\prime}
                P(k, k^\prime) \log P(k, k^\prime)
            P (k, k^\prime) = \frac{
                \left| C_k \cap C_{k^\prime}^\prime \right|}{n}

        The joint entropy is minimized when there is a one-to-one mapping
        between the pixels in C and their counterparts in C^\prime and it
        increases as the statistical relationship between C and C^\prime
        weakens.

    Parameters
    ----------
    probs : torch.Tensor
        A segmentation or clustering C with K segments with shape N x K x SP
    probs_prime : torch.Tensor
        A segmentation or clustering C^\prime with K^\prime segments with shape
        N x K^\prime x SP.
    smooth : float, optional
        Soften the normalized result. The default is 10e-8.

    Returns
    -------
    joint_entropy : torch.Tensor
        The joint entropy is a tensor of shape N.

    """
    num_elements = torch.prod(torch.tensor(probs.shape[2:]))

    joint_probability_distribution = (
        contingency_table_fn(probs, probs_prime) / num_elements
    )

    joint_entropy = -torch.sum(
        joint_probability_distribution
        * torch.log(joint_probability_distribution.clamp(min=smooth)),
        dim=(1, 2),
    )

    return joint_entropy


def mutual_information(
    output: torch.Tensor,
    target: torch.Tensor,
    labels: torch.Tensor = None,
    normalization: bool = False,
    reduction: str = "mean",
    smooth: float = 10e-6,
) -> torch.Tensor:
    r"""
    Estimate mutual information for a discrete target variable.

    Notes
    -----
        The mutual information between the clusterings/segmentations C,
        C^\prime to be equal to the mutual information between the associated
        random variables

        .. math:

            MI (C, C^\prime) & = \sum_{k=1}^K \sum_{k^\prime = 1}^{K^\prime}
                P (k, k^\prime) \log \frac{P (k, k^\prime)}
                {P(k) P^\prime (k^\prime)} \\
                & = \sum_{k=1}^{K} \sum_{k^\prime = 1}^{K^\prime}
                    \frac{n_{k, k^\prime}}{n}
                    \log{
                        \frac{n_{k, k^\prime}}{n}
                        \frac{n_{k}}{n}
                        \frac{n_{k^\prime}}{n}
                    } \\
                & = H(C) + H(C^\prime) - H(C, C^\prime)

        It measures the amount of information one variable has about the other.
        The more the segmented result resembles the grount truth, the more
        information is shared.

    Parameters
    ----------
    output : torch.Tensor
        tensor with labels one-hot encoded in dim 1 and batch size in dim 0.
    target : torch.Tensor
        tensor with labels one-hot encoded in dim 1 and batch size in dim 0.
    labels : torch.Tensor, optional
        labels to include in MI. The default is None.
    normalization : bool, optional
        Should MI be normalized by geometric mean of the entropies after Strehl
        & Ghosh since it has no argument-independent upper bound. The default
        is False.
    reduction : str, optional
        Modality how to reduce result over samples. The default is "mean".
    smooth : float, optional
        Soften the normalized result. The default is 10e-6.

    Raises
    ------
    ValueError
        If reduction is not available.

    Returns
    -------
    mi : torch.Tensor
        Mutual information.

    References
    ----------
        [1] P. Viola and W. M. Wells, "Alignment by maximization of mutual
            information," Proceedings of IEEE International Conference on
            Computer Vision, 1995, pp. 16-23, doi: 10.1109/ICCV.1995.466930.
            https://ieeexplore.ieee.org/document/466930

    """
    if reduction not in ["mean", "sum", "none"]:
        raise ValueError(f"Reduction {reduction} is not available.")

    if labels is not None:
        output = torch.index_select(output, dim=1, index=labels)
        target = torch.index_select(target, dim=1, index=labels)

    entropy_output = marginal_entropy_fn(output)
    entropy_target = marginal_entropy_fn(target)
    joint_entropy = joint_entropy_fn(output, target)

    mi = entropy_output + entropy_target - joint_entropy

    if normalization:
        mi = mi / torch.sqrt(entropy_output * entropy_target).clamp(min=smooth)

    if reduction == "mean":
        mi = torch.nanmean(mi)
    elif reduction == "sum":
        mi = torch.nansum(mi)

    return mi


def variation_of_information(
    output: torch.Tensor,
    target: torch.Tensor,
    labels: torch.Tensor = None,
    normalization: bool = False,
    reduction: str = "mean",
) -> torch.Tensor:
    r"""
    Comparing segmentations by calculating variation of information.

    Notes
    -----
        The variation of information represents the distance between two
        segmentations as the average conditional entropy of one segmentation
        given the other, and as a result measures the amount of randomness in
        one segmentation which cannot be explained by the other.

        The variation of information (dVI ), introduced in _[1] with properties
        _[2], measures the distance between two clusterings of a data set D in
        terms of the information difference between them.

        .. math::

            d_{VOI}(C, C^\prime) & = H(C) + H(C^\prime) - 2 I (C, C^\prime) \\
                & = 2 H (C, C^\prime) - H(C) - H(C^\prime)

        where H and I represent respectively the entropies of and the mutual
        information between the two clusterings, n is the number of data
        points.

    Parameters
    ----------
    output : torch.Tensor
        tensor with labels one-hot encoded in dim 1 and batch size in dim 0.
    target : torch.Tensor
        tensor with labels one-hot encoded in dim 1 and batch size in dim 0.
    labels : torch.Tensor, optional
        labels to include in MI. The default is None.
    normalization : bool, optional
        Normalizing by log n. Not recommended if we are to make comparisons
        between distances obtained on different sizes. The default is False.
    reduction : str, optional
        Modality how to reduce result over samples. The default is "mean".

    Raises
    ------
    ValueError
        If reduction is not available.

    Returns
    -------
    voi : torch.Tensor
        Variation of information.

    References
    ----------
        .. [1] Meilă, M. (2003). Comparing clusterings by the variation of
            information. Proceedings of the Sixteenth Annual Conference of
            Computational Learning Theory (COLT). Springer.
            https://rd.springer.com/content/pdf/10.1007%2Fb12006.pdf
        .. [2] Meilă, M. (2005). Comparing clusterings: An axiomatic view.
            Proceedings of the 22nd International Conference on Machine
            learning. 577-584. 10.1145/1102351.1102424.
            https://sites.stat.washington.edu/mmp/Papers/icml05-compare-axioms.pdf

    """
    if reduction not in ["mean", "sum", "none"]:
        raise ValueError(f"Reduction {reduction} is not available.")

    if labels is not None:
        raise NotImplementedError(
            "Reduction to certain labels is not implemented yet.")
        # output = torch.index_select(output, dim=1, index=labels)
        # target = torch.index_select(target, dim=1, index=labels)

    entropy_output = marginal_entropy_fn(output)
    entropy_target = marginal_entropy_fn(target)
    joint_entropy = joint_entropy_fn(output, target)

    voi = 2 * joint_entropy - entropy_output - entropy_target

    if normalization:
        num_elements = torch.prod(torch.tensor(output.shape[2:]))
        voi = voi / torch.log(num_elements)

    if reduction == "mean":
        voi = torch.nanmean(voi)
    elif reduction == "sum":
        voi = torch.nansum(voi)

    return voi
