#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The module provides tensor preparation functionality to compute metrices.

Created on Thu Mar 10 11:43:14 2022

@author: jsteiglechner
"""

from typing import Tuple

import torch
import torch.nn.functional as F


def get_tensor_label_type(tensor: torch.Tensor) -> str:
    """
    Get label type of tensor.

    Parameters
    ----------
    tensor : torch.Tensor
        tensor to evaluate.

    Returns
    -------
    str
        type of labels.

    """
    labels = torch.unique(tensor)

    if len(labels) < 3:
        return "binary"
    elif torch.all((labels // 1) == 0):
        return "multiclass"
    else:
        return "soft"


def get_tensor_difference(
        tensor_a: torch.Tensor, tensor_b: torch.Tensor) -> int:
    """
    Get the values of one 1-d tensor that are not in a second 1-d tensor.

    Parameters
    ----------
    tensor_a : torch.Tensor
        Some tensor.
    tensor_b : torch.Tensor
        Some other tensor.

    Returns
    -------
    int
        DESCRIPTION.

    """
    return tensor_a[~ tensor_a.unsqueeze(1).eq(tensor_b).any(-1)]


def unique_labels(output: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
    """
    Determine unique labels.

    Parameters
    ----------
    output : torch.Tensor
        model output.
    target : torch.Tensor
        true values.

    Returns
    -------
    labels : torch.Tensor or None
        label set.

    """
    target_values = torch.unique(target)

    label_type_output = get_tensor_label_type(output)
    label_type_target = get_tensor_label_type(target)

    if label_type_target == "binary":
        labels = target_values
    elif (label_type_target == "multiclass") and (label_type_output == "soft"):
        if output.shape == target.shape:
            labels = torch.arange(
                torch.max(target_values),
                requires_grad=False,
                dtype=torch.int,
                device=target.device,
            )
        else:
            labels = torch.arange(
                get_tensor_difference(
                    torch.tensor([*output.shape], dtype=torch.int),
                    torch.tensor([*target.shape], dtype=torch.intint))[0],
                requires_grad=False,
                dtype=torch.int,
                device=target.device,
            )
    else:
        labels = None

    return labels


def encode_to_batch_times_class_times_spatial(
        tensor: torch.Tensor,
        num_classes: int,
        spatial_dim: int = 3) -> torch.Tensor:
    """
    Encode tensor to match shape [N x C x R].

    N for size of batch, C for number of classes, R for spatial expansion.

    Parameters
    ----------
    tensor : torch.Tensor
        tensor with shape [R] or [C, R] if N == 1, [N, R], or [N, C, R] else.
    num_classes : int
        number of classes C.
    spatial_dim : int, optional
        number of spatial dimensions. The default is 3.

    Raises
    ------
    ValueError
        If spatial dimensionality is too low.
        If spatial dimensionality is too high.

    Returns
    -------
    tensor_encoded : torch.Tensor
        Tensor with shape [N x C x spatial expansion].

    """
    device = tensor.device

    if tensor.dim() < spatial_dim:
        raise ValueError('Tensor spatial dimensionality is too low.')
    if tensor.dim() > spatial_dim + 2:
        raise ValueError('Tensor spatial dimensionality is too high.')

    if tensor.dim() == spatial_dim:
        tensor = tensor.unsqueeze(0).to(device)

    tensor_shape = list(tensor.shape)  # [N, R]
    if tensor.dim() == spatial_dim + 1:
        if tensor_shape[0] == num_classes:
            tensor_encoded = tensor.unsqueeze(0).to(device)
        else:
            tensor_encoded = torch.zeros([tensor_shape[0], num_classes] +
                                         tensor_shape[-spatial_dim:]).to(
                device)
            tensor_encoded.scatter_(1, tensor.long().unsqueeze(1), 1)
    elif tensor.dim() == spatial_dim + 2:
        if tensor_shape[1] == num_classes:
            tensor_encoded = tensor.clone()
        else:
            tensor_encoded = torch.zeros([tensor_shape[0], num_classes] +
                                         tensor_shape[-spatial_dim:]).to(
                device)
            tensor_encoded.scatter_(1, tensor.long(), 1)

    return tensor_encoded


def get_structuring_kernel(
        num_dim: int = 3,
        pixel_distance: int = 1,
        kernel_type: str = 'box',
        device: torch.device = torch.device('cpu')) -> torch.Tensor:
    """
    Return a kernel for morphological operations.

    The function constructs that can be used as kernel for operations like
    erode or dilate.

    Parameters
    ----------
    num_dim : int, optional
        Dimensionality of structuring element. The default is 3.
    pixel_distance : int, optional
        Size of the structuring element. The default is 1.
    kernel_type : str, optional
        Shape of kernel. The default is 'box'.
    device : torch.device, optional
        Device where to store tensor. The default is torch.device('cpu').

    Raises
    ------
    ValueError
        If kernel_type is not supported.

    Returns
    -------
    kernel_tensor : torch.Tensor
        Binary kernel.

    """
    if kernel_type not in ['box', 'cross', 'l1_sphere', 'sphere']:
        raise ValueError(f"Erosion type {kernel_type} is not supported.")

    kernel_size = 2 * pixel_distance + 1

    if kernel_type == 'box':
        kernel_tensor = torch.ones(
            tuple(kernel_size for i in range(num_dim)),
            dtype=int,
            requires_grad=False,
            device=device,
        )

    elif kernel_type == 'cross':
        grid = torch.meshgrid(
            *[torch.arange(kernel_size) for dim in range(num_dim)],
            indexing='xy',
        )
        kernel_tensor = torch.any(
            torch.cat(
                [grid_i[None] for grid_i in grid], dim=0) == pixel_distance,
            dim=0,
        ).to(dtype=int, device=device)

    elif kernel_type == 'l1_sphere':
        grid = torch.meshgrid(
            *[torch.arange(kernel_size) for dim in range(num_dim)],
            indexing='xy',
        )
        distance_from_center = torch.sum(torch.cat(
            [torch.abs(grid_i - pixel_distance)[None] for grid_i in grid],
            dim=0),
            dim=0,
        )

        kernel_tensor = (distance_from_center <= pixel_distance).to(
            dtype=int, device=device)

    elif kernel_type == 'sphere':
        grid = torch.meshgrid(
            *[torch.arange(kernel_size) for dim in range(num_dim)],
            indexing='xy',
        )
        distance_from_center = torch.sqrt(
            torch.sum(torch.cat(
                [((grid_i - pixel_distance)**2)[None] for grid_i in grid],
                dim=0),
                dim=0,
            )
        )

        kernel_tensor = (distance_from_center <= pixel_distance).to(
            dtype=int, device=device)

    return kernel_tensor


def check_boundary_kernel_fullness(
        mask: torch.Tensor,
        boundary_width: int = 1,
        kernel_type: str = 'box',
) -> Tuple[torch.Tensor, int]:
    """
    Calculate the fullness of a pixel kernel and its maximum value.

    Parameters
    ----------
    mask : torch.Tensor
        Source image.
    boundary_width : int, optional
        Size of the structuring element. The default is 1.
    kernel_type : str, optional
        Shape of kernel. The default is 'box'.

    Returns
    -------
    kernel_fullness : torch.Tensor
        Tensor that describes how full is the kernel at a certain pixel.
    threshold_kernel_full : int
        Number of elements so that the kernel is full.

    """
    kernel_tensor = get_structuring_kernel(
        num_dim=len(mask.shape[2:]),
        pixel_distance=boundary_width,
        kernel_type=kernel_type,
        device=mask.device,
    )

    threshold_kernel_full = kernel_tensor.sum()
    kernel_tensor = kernel_tensor.unsqueeze(
        0).unsqueeze(0).repeat(mask.shape[1], 1, 1, 1, 1)

    if (torch.min(mask) >= 0.) and (torch.max(mask) <= 1.):
        data_type = torch.half
    elif threshold_kernel_full < 256:
        data_type = torch.uint8
    else:
        data_type = torch.int16

    kernel_fullness = F.conv3d(
        mask.to(float),
        kernel_tensor.to(float),
        padding='same',
        groups=mask.shape[1],
    )
    kernel_fullness = kernel_fullness.to(data_type)

    return kernel_fullness, threshold_kernel_full


def erode_labels(
        mask: torch.Tensor,
        erosion_distance: int = 1,
        erosion_type: str = 'box',
) -> torch.Tensor:
    """
    Perform the erosian operation on mask.

    Parameters
    ----------
    mask : torch.Tensor
        The source image.
    erosion_distance : int, optional
        Size of kernel we will use to perform the operation. The default is 1.
    erosion_type : str, optional
        Type of kernel we will use. The default is 'box'.

    Returns
    -------
    eroded_mask : torch.Tensor
        Eroded mask.

    """
    kernel_fullness, threshold_kernel_full = check_boundary_kernel_fullness(
        mask=mask,
        boundary_width=erosion_distance,
        kernel_type=erosion_type
    )
    eroded_mask = (kernel_fullness == threshold_kernel_full).to(mask.dtype)

    return eroded_mask


def erode_labels_complement(
        mask: torch.Tensor,
        erosion_distance: int = 1,
        erosion_type: str = 'box',
) -> torch.Tensor:
    """
    Perform the erosian operation complementary on mask.

    Parameters
    ----------
    mask : torch.Tensor
        The source image.
    erosion_distance : int, optional
        Size of kernel we will use to perform the operation. The default is 1.
    erosion_type : str, optional
        Type of kernel we will use. The default is 'box'.

    Returns
    -------
    complementary_eroded_mask : torch.Tensor
        Complementarily eroded mask.

    """
    kernel_fullness, threshold_kernel_full = check_boundary_kernel_fullness(
        mask=mask,
        boundary_width=erosion_distance,
        kernel_type=erosion_type
    )
    complementary_eroded_mask = (
        kernel_fullness < threshold_kernel_full).to(mask.dtype)

    return complementary_eroded_mask


def mask_to_boundary_region(
        mask: torch.Tensor,
        boundary_width: int = 1,
        kernel_type: str = 'box',
        lower_threshold_num_elements: int = 1,
) -> torch.Tensor:
    """
    Return set of pixels in the boundary region of a binary mask.

    The boundary is a region of pixels within pixel distance from the contour
    pixels.

    Parameters
    ----------
    mask : torch.Tensor
        Source image.
    boundary_width : int, optional
        Pixel distance. The default is 1.
    kernel_type : str, optional
        Type of kernel to use for boundary extraction. The default is 'box'.
    threshold_num_elements: int, optional
        Threshold for number of elements required for not being background.
        Default is 1.

    Returns
    -------
    boundary_region: torch.Tensor
        Set of pixels in the boundary region of a binary mask.

    """
    kernel_fullness, threshold_kernel_full = check_boundary_kernel_fullness(
        mask=mask,
        boundary_width=boundary_width,
        kernel_type=kernel_type
    )
    boundary_region = (
        (kernel_fullness < threshold_kernel_full)
        * (kernel_fullness >= lower_threshold_num_elements)
    ).to(mask.dtype)

    return boundary_region
