#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
The module calculates spatial distance based scores.

Created on Thu Mar 17 14:42:04 2022

@author: jsteiglechner
"""
import psutil
import warnings

import numpy as np
import torch

from .prepare_tensors import mask_to_boundary_region


def euclidean_distance_to_point_cloud(
    positions_x: torch.Tensor,
    positions_y: torch.Tensor,
    directed: bool = True,
) -> torch.Tensor:
    """
    Calculate the euclidean distance from a point to a point cloud.

    Parameters
    ----------
    positions_x : torch.Tensor
        Position of point.
    pos_y : torch.Tensor
        Positions of points in point cloud.

    Returns
    -------
    torch.Tensor
        Minimal euclidean distance.

    """
    if directed:
        distances = squared_euclidean_distance_to_point_cloud(
            positions_x,
            positions_y,
            directed,
        )
        return torch.sqrt(distances)
    else:
        distances_a, distances_b = squared_euclidean_distance_to_point_cloud(
            positions_x,
            positions_y,
            directed,
        )
        return torch.sqrt(distances_a), torch.sqrt(distances_b)


def squared_euclidean_distance_to_point_cloud(
    positions_x: torch.Tensor,
    positions_y: torch.Tensor,
    directed: bool = True,
) -> torch.Tensor:
    """
    Calculate the squared euclidean distance from a point to a point cloud.

    Parameters
    ----------
    pos_x : torch.Tensor
        Positions of points in source cloud.
    pos_y : torch.Tensor
        Positions of points in target cloud.

    Returns
    -------
    torch.Tensor
        Minimal euclidean distance.

    """
    if positions_x.dtype != torch.int32:
        warnings.warn(
            "Data type of positions is not int32. This can lead to mistakes.",
        )
    if positions_y.dtype != torch.int32:
        warnings.warn(
            "Data type of positions is not int32. This can lead to mistakes.",
        )

    distances = torch.empty(
        len(positions_x),
        dtype=torch.int32,
        device=positions_x.device,
        requires_grad=positions_x.requires_grad,
    )
    if not directed:
        o_distances = (
            torch.ones(
                len(positions_y),
                dtype=torch.int32,
                device=positions_x.device,
                requires_grad=positions_x.requires_grad,
            )
            * (2 ** 31 - 1)
        )

    if positions_x.device.type == "cuda":
        batch_size = torch.cuda.get_device_properties(
            positions_x.device
        ).multi_processor_count
    else:
        batch_size = psutil.cpu_count()

    num_iterations = np.ceil(len(positions_x) / batch_size)
    for i_batch in torch.arange(num_iterations):
        start_idx = int(i_batch * batch_size)
        end_idx = min(int((i_batch + 1) * batch_size), len(positions_x))

        differences = torch.sub(
            positions_x[start_idx:end_idx, None],
            positions_y[None],
        )
        differences = torch.pow(differences, 2)
        differences = torch.sum(differences, dim=2)

        distances[start_idx:end_idx] = torch.min(differences, dim=1).values

        if not directed:
            o_distances = torch.minimum(
                o_distances, torch.min(differences, dim=0).values
            )

    if directed:
        return distances.to(torch.float32)
    else:
        return distances.to(torch.float32), o_distances.to(torch.float32)


def measure_point_cloud_distance(
    positions_x: torch.Tensor,
    positions_y: torch.Tensor,
    distance_function: str,
    directed: bool = False,
) -> torch.Tensor:
    """
    Measure the distance of one point to a cloud of points.

    Parameters
    ----------
    pos_x : torch.Tensor
        Positions of points in source cloud.
    pos_y : torch.Tensor
        Positions of points in target cloud.
    distance_function : str
        Metric to use for distance calculation.
    directed : bool, optional
        DESCRIPTION. The default is False.

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    torch.Tensor
        Distance to point cloud.

    """
    if positions_x.dim() != positions_y.dim():
        raise ValueError("Dimensionality of point clouds do not match.")

    if distance_function == "euclidean":
        distances = euclidean_distance_to_point_cloud(
            positions_x,
            positions_y,
            directed=directed,
        )
    elif distance_function == "squared_euclidean":
        distances = squared_euclidean_distance_to_point_cloud(
            positions_x, positions_y, directed=directed, )
    else:
        raise ValueError(
            f"Distance function {distance_function} is not supperted."
        )

    return distances


def measure_percentile_binary_hausdorff_distance(
    surface_a: torch.Tensor,
    surface_b: torch.Tensor,
    distance_function: str = "euclidean",
    percentile: int = None,
    directed: bool = True,
) -> torch.Tensor:
    r"""
    Compute the directed binary Hausdorff distance at a certain percentile.

    Parameters
    ----------
    surface_a : torch.Tensor
        Input surface.
    surface_b : torch.Tensor
        Reference surface.
    distance_function : str, optional
        Specifies the metric. The default is 'euclidean'.
    percentile : int, optional
        Specifies the percentile of Hausdorff distance. The default is None.
    directed : bool, optional
        DESCRIPTION. The default is False.

    Raises
    ------
    ValueError
        If percentile is out of allowed range.

    Returns
    -------
    directed_hausdorff_distance : torch.Tensor
        result.

    """
    surface_a_indices = surface_a.nonzero(as_tuple=False).to(torch.int32)
    surface_b_indices = surface_b.nonzero(as_tuple=False).to(torch.int32)

    distances = measure_point_cloud_distance(
        positions_x=surface_a_indices,
        positions_y=surface_b_indices,
        distance_function=distance_function,
        directed=directed,
    )

    if (not percentile) or (percentile == 100):
        hausdorff_distance = torch.max(distances)
    elif not (0 < percentile < 100):
        raise ValueError(
            "Percentile has to be a value between 0 and 100, "
            f"got {percentile}. Take maximum value."
        )
        hausdorff_distance = torch.max(distances)
    elif directed:
        hausdorff_distance = torch.quantile(
            distances,
            q=percentile / 100,
        )
    else:
        hausdorff_distance_a = torch.quantile(
            distances[0],
            q=percentile / 100,
        )
        hausdorff_distance_b = torch.quantile(
            distances[1],
            q=percentile / 100,
        )
        hausdorff_distance = torch.max(
            hausdorff_distance_a, hausdorff_distance_b)

    return hausdorff_distance


def hausdorff_distance_fn(
    output: torch.Tensor,
    target: torch.Tensor,
    distance_function: str = "euclidean",
    percentile: int = None,
    directed: bool = False,
) -> torch.Tensor:
    r"""
    Measures how far two subsects of a metric space are from each other.

    Notes
    -----
        Computes the Hausdorff distance between the set of non-zero pixels of
        two images. We implement a torch-based version of [1]_. It is based on
        the directed Hausdorff distance

        .. math::

            d (a, B) = \inf_{b \in B} d(a, b) \\
            d_H (X, Y) = \max \left{ \sup_{x \in X} d(x, Y),
                                     \sup_{y \in Y} d(X, y) \right}

    Parameters
    ----------
    output : torch.Tensor
        tensor with labels one-hot encoded. Shape is N x C x Spatial.
    target : torch.Tensor
        tensor with labels one-hot encoded. Shape is N x C x Spatial.
    distance_function : str, optional
        DESCRIPTION. The default is 'euclidean'.
    percentile : int, optional
        DESCRIPTION. The default is None.
    directed : bool, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    hausdorff_distance : torch.Tensor
        Calculated Hausdorff distance.

    References
    ----------
        .. [1] A. A. Taha and A. Hanbury, "An Efficient Algorithm for
            Calculating the Exact Hausdorff Distance," in IEEE Transactions on
            Pattern Analysis and Machine Intelligence, vol. 37, no. 11,
            pp. 2153-2163, 1 Nov. 2015, doi: 10.1109/TPAMI.2015.2408351.

    """
    (batch_size, num_classes) = target.shape[:2]
    hausdorff_distance = torch.empty(
        (batch_size, num_classes), device=target.device, dtype=torch.half
    )

    for i_batch, i_label in torch.cartesian_prod(
        torch.arange(batch_size, device=target.device),
        torch.arange(num_classes, device=target.device),
    ):
        selected_output = (
            torch.index_select(
                torch.index_select(output, dim=1, index=i_label),
                dim=0,
                index=i_batch,
            )
            .squeeze()
            .squeeze()
        )
        selected_target = (
            torch.index_select(
                torch.index_select(target, dim=1, index=i_label),
                dim=0,
                index=i_batch,
            )
            .squeeze()
            .squeeze()
        )

        # Check that labels exist in botch images.
        if not torch.any(selected_output):
            warnings.warn(f"The cardinality of label {i_label} output is 0.")
            hausdorff_distance[
                i_batch,
                i_label,
            ] = torch.nan
        elif not torch.any(selected_target):
            warnings.warn(
                f"The cardinality of label {i_label} in target is 0.")
            hausdorff_distance[
                i_batch,
                i_label,
            ] = torch.nan
        else:
            hausdorff_distance[
                i_batch,
                i_label,
            ] = measure_percentile_binary_hausdorff_distance(
                selected_output,
                selected_target,
                distance_function,
                percentile,
                directed,
            )

    return hausdorff_distance


def hausdorff_distance_metric(
    output: torch.Tensor,
    target: torch.Tensor,
    distance_function: str = "euclidean",
    percentile: int = None,
    directed: bool = False,
    labels: torch.Tensor = None,
    pos_label: int = 1,
    label_reduction: str = "mean",
    reduction: str = "mean",
) -> torch.Tensor:
    """
    Calculate the hausdorff distance metric of output and target.

    Parameters
    ----------
    output : torch.Tensor
        tensor with labels one-hot encoded in pos_label.
    target : torch.Tensor
        tensor with labels one-hot encoded in pos_label.
    distance_function : str, optional
        Specifies the metric. The default is 'euclidean'.
    percentile : int, optional
        Specifies the percentile of Hausdorff distance. The default is None.
    directed : bool, optional
        Specifies directed Hausdorff distance. The default is False.
    labels : torch.Tensor, optional
        Labels to include in IoU. The default is None.
    pos_label : int, optional
        Specifies in which dim labels are specified. The default is 1.
    label_reduction : str, optional
        Modality how to reduce result over labels. The default is 'mean'.
    reduction : str, optional
        Modality how to reduce result over samples. The default is 'mean'.

    Raises
    ------
    ValueError
        If reduction is not available.

    Returns
    -------
    hd : torch.Tensor
        Hausdorff distance.

    """
    if reduction not in ["mean", "sum", "none"]:
        raise ValueError(f"Reduction {reduction} is not available.")
    if label_reduction not in ["mean", "sum", "max", "none"]:
        raise ValueError(f"Reduction {label_reduction} is not available.")

    if pos_label != 1:
        output = torch.movedim(output, pos_label, 1)
        target = torch.movedim(target, pos_label, 1)

    if labels is not None:
        output = torch.index_select(output, dim=pos_label, index=labels)
        target = torch.index_select(target, dim=pos_label, index=labels)

    output = (
        mask_to_boundary_region(
            mask=output,
            boundary_width=1,
            kernel_type="cross",
            lower_threshold_num_elements=1,
        )
        * output
    )
    target = (
        mask_to_boundary_region(
            mask=target,
            boundary_width=1,
            kernel_type="cross",
            lower_threshold_num_elements=1,
        )
        * target
    )

    hausdorff_distance = hausdorff_distance_fn(
        output,
        target,
        percentile=percentile,
        directed=directed,
    )

    # reduction over labels
    if label_reduction == "mean":
        hausdorff_distance = torch.nanmean(hausdorff_distance, dim=1)
    elif label_reduction == "sum":
        hausdorff_distance = torch.nansum(hausdorff_distance, dim=1)
    elif label_reduction == "max":
        hausdorff_distance = torch.nanmax(hausdorff_distance, dim=1)

    # reduction over samples
    if reduction == "mean":
        hausdorff_distance = torch.nanmean(hausdorff_distance, dim=0)
    elif reduction == "sum":
        hausdorff_distance = torch.nansum(hausdorff_distance, dim=0)

    return hausdorff_distance
