"""
The module calculates spatial overlap based scores.

Created on Mon Mar  7 16:51:41 2022

@author: jsteiglechner
"""

from typing import List

import torch


from .prepare_tensors import (
    encode_to_batch_times_class_times_spatial, mask_to_boundary_region)


def accuracy(output: torch.Tensor, target: torch.Tensor):
    r"""
    Check how close output is to target.

    Assign a max score of 1 if the voxel is correctly predicted, 0 otherwise.

    .. math::

        \frac{1}{N} \sum_{i=0}^N 1_{(X_i = Y_i)}

    Parameters
    ----------
    output : torch.Tensor
        prediction of a model.
    target : torch.Tensor
        true values.

    Returns
    -------
    accuracy : torch.Tensor
        accuracy.

    """
    tmp_output = output.flatten()
    tmp_target = target.flatten()

    accuracy = (tmp_output * tmp_target).sum() / torch.numel(target)

    return accuracy


def dice_coefficient(output: torch.Tensor,
                     target: torch.Tensor,
                     smooth: float = 10e-5) -> torch.Tensor:
    r"""
    Compute the Dice dissimilarity regarding a boolean tensor.

    The Dice dissimilarity between `X` and `Y` is

    .. math::

        \frac{2 | X \cap Y |}{|X| + |Y|}

    Parameters
    ----------
    output : torch.Tensor
        prediction of a model. Values can be discrete or contigious.
    target : torch.Tensor
        true values. They should be binary.
    smooth : float, optional
        softens the result. The default is 10e-5.

    Returns
    -------
    dice_score : torch.Tensor
        Dice similarity coefficient.

    """
    # Note, that .flatten is equivalent to .contiguous().view(-1)
    tmp_output = output.flatten()
    tmp_target = target.flatten()
    intersection = (tmp_output * tmp_target).sum()

    dice_score = ((2. * intersection).clamp(min=smooth) /
                  (tmp_output.sum() + tmp_target.sum()).clamp(min=smooth))

    return dice_score


def continous_dice_coefficient(output: torch.Tensor,
                               target: torch.Tensor,
                               smooth: float = 10e-5) -> torch.Tensor:
    r"""
    Extend the definition of the classical Dice coefficient.

    See paper: https://www.biorxiv.org/content/10.1101/306977v1

    .. math::

        \frac{2 | X \cap Y |}{c |X| + |Y|} \\

    Parameters
    ----------
    output : torch.Tensor
        DESCRIPTION.
    target : torch.Tensor
        DESCRIPTION.
    smooth : float, optional
        DESCRIPTION. The default is 10e-5.

    Returns
    -------
    continous_dice_score : TYPE
        DESCRIPTION.

    """
    tmp_output = output.flatten()
    tmp_target = target.flatten()
    intersection = (tmp_output * tmp_target).sum()

    continous_factor = (tmp_output * torch.sign(tmp_target)).sum()

    continous_dice_score = (
        (2. * intersection).clamp(min=smooth) /
        (continous_factor * tmp_output.sum()
         + tmp_target.sum()).clamp(min=smooth)
    )

    return continous_dice_score


def dice_score_per_class(output: torch.Tensor,
                         target: torch.Tensor,
                         class_dim: int = 1,
                         smooth: float = 10e-5) -> torch.Tensor:
    r"""
    Calculate dice score per class.

    Parameters
    ----------
    output : torch.Tensor
        tensor with classes one-hot encoded in class_dim.
    target : torch.Tensor
        tensor with classes one-hot encoded in class_dim.
    class_dim : int, optional
        specifies in which dim classes are specified. The default is 1.
    smooth : float, optional
        soften the result. The default is 10e-5.

    Raises
    ------
    ValueError
        If shape of output and target do not match.

    Returns
    -------
    dice_score_per_class : torch.Tensor
        dices of shape of num classes C.

    """
    if target.shape != output.shape:
        raise ValueError("Target shape and output shape do not match.")

    dims = torch.arange(output.dim())
    dims = tuple(dims[dims != class_dim])

    output_sum_per_class = output.sum(dim=dims)
    target_sum_per_class = target.sum(dim=dims)
    intersection_per_class = (output * target).sum(dim=dims)
    dice_score_per_class = (
        (2. * intersection_per_class).clamp(min=smooth) /
        (output_sum_per_class + target_sum_per_class).clamp(min=smooth))

    return dice_score_per_class


def generalized_dice_score(output: torch.Tensor,
                           target: torch.Tensor,
                           num_classes: int,
                           classes_to_ignore: List = [],
                           spatial_dim: int = 3) -> torch.Tensor:
    """
    Measure overlap of labelled regions by label.

    There are several variations of generalized Dice scores. Our function
    follows Crum et al. https://ieeexplore.ieee.org/document/1717643

    Parameters
    ----------
    output : torch.Tensor
        model output logits.
    target : torch.Tensor
        true values.
    num_classes : int
        number of classes to predict.
    classes_to_ignore : List, optional
        Classes which can be ignored. The default is [].
    spatial_dim : int, optional
        spatial dimensionality. The default is 3.

    Raises
    ------
    ValueError
        If target and output do not have the same spatial expansion..

    Returns
    -------
    dice_score_per_class: torch.Tensor
        dices with shape [num_classes - len(classes_to_ignore)].

    """
    if output.shape[-spatial_dim:] != target.shape[-spatial_dim:]:
        raise ValueError(
            "Target and output do not have the same spatial expansion.")

    # Fit target to B x C x spatial
    output_encoded = encode_to_batch_times_class_times_spatial(
        output, num_classes, spatial_dim)
    target_encoded = encode_to_batch_times_class_times_spatial(
        target, num_classes, spatial_dim)

    if classes_to_ignore != []:
        classes = torch.arange(num_classes)
        classes_index = torch.all(
            torch.stack(
                [classes != class_to_ignore
                 for class_to_ignore in classes_to_ignore]), dim=0)
        output_encoded = output_encoded[:, classes_index]
        target_encoded = target_encoded[:, classes_index]

    return dice_score_per_class(output_encoded, target_encoded, class_dim=1)


def binary_intersection_over_union(
        output: torch.Tensor,
        target: torch.Tensor,
        smooth: float = 10e-6,
) -> torch.Tensor:
    """
    Measure th similarity of (soft) single-class tensors.

    Parameters
    ----------
    output : torch.Tensor
        tensors with values between 0 and 1.
    target : torch.Tensor
        tensors with values between 0 and 1.
    smooth : float, optional
        soften the result. The default is 10e-6.

    Returns
    -------
    jaccard_index : TYPE
        Intersection over Union.

    """
    tmp_output = output.flatten()
    tmp_target = target.flatten()
    intersection = (tmp_output * tmp_target).sum()

    jaccard_index = (
        intersection.clamp(min=smooth) /
        (tmp_output.sum() + tmp_target.sum() - intersection).clamp(min=smooth)
    )

    return jaccard_index


def intersection_over_union(
        output: torch.Tensor,
        target: torch.Tensor,
        labels: torch.Tensor = None,
        pos_label: int = 1,
        batch_iou: bool = False,
        non_presence_threshold: float = None,
        reduction: str = 'mean',
        smooth: float = 10e-6,
) -> torch.Tensor:
    r"""
    Measure the similarity of output and target.

    Parameters
    ----------
    output : torch.Tensor
        tensor with labels one-hot encoded in pos_label.
    target : torch.Tensor
        tensor with labels one-hot encoded in pos_label.
    labels : torch.Tensor, optional
        labels to include in IoU. The default is None.
    pos_label : int, optional
        specifies in which dim labels are specified. The default is 1.
    batch_iou : bool, optional
        Calculate IoU for each sample. The default is False.
    non_presence_threshold : float, optional
        Threshold at which number of pixels to include. The default is None.
    reduction : str, optional
        Modality how to reduce result. The default is 'mean'.
    smooth : float, optional
        soften the result. The default is 10e-6.

    Raises
    ------
    ValueError
        If reduction is not available.

    Returns
    -------
    jaccard_index : torch.Tensor
        Intersection over Union. NaN if classes are not presence.

    """
    if reduction not in ['mean', 'sum', 'none']:
        raise ValueError(f"Reduction {reduction} is not available.")

    if labels is not None:
        output = torch.index_select(output, dim=pos_label, index=labels)
        target = torch.index_select(target, dim=pos_label, index=labels)

    dims = torch.arange(output.dim())
    dims = tuple(dims[dims != pos_label])

    if batch_iou:
        dims = dims[1:]

    output_sum_per_class = output.sum(dim=dims)
    target_sum_per_class = target.sum(dim=dims)
    intersection_per_class = (output * target).sum(dim=dims)

    jaccard_index = (
        intersection_per_class.clamp(min=smooth)
        / (
            output_sum_per_class
            + target_sum_per_class
            - intersection_per_class
        ).clamp(min=smooth)
    )

    if non_presence_threshold is not None:
        jaccard_index[
            (target_sum_per_class + output_sum_per_class)
            <= non_presence_threshold
        ] = torch.nan

    if reduction == 'mean':
        jaccard_index = torch.nanmean(jaccard_index)
    elif reduction == 'sum':
        jaccard_index = torch.nansum(jaccard_index)

    return jaccard_index


def boundary_intersection_over_union(
    output: torch.Tensor,
    target: torch.Tensor,
    boundary_width: int = None,
    kernel_type: str = 'box',
    labels: torch.Tensor = None,
    pos_label: int = 1,
    batch_iou: bool = False,
    non_presence_threshold: float = None,
    reduction: str = 'mean',
    smooth: float = 10e-6,
) -> torch.Tensor:
    r"""
    Measure the similarity of output boundary and target boundary.

    Notes
    -----
        The boundary_intersection_over_union is a torch-based implementation of
        the boundery IoU [1]_

        .. math::

            \text{Boundary IoU}(G, P) = \frac{(G_d \cap G) \cap (P_d \cap P)}
                {(G_d \cap G) \cup (P_d \cap P)}

    Parameters
    ----------
    output : torch.Tensor
        tensor with labels one-hot encoded in pos_label.
    target : torch.Tensor
        tensor with labels one-hot encoded in pos_label.
    boundary_width : int, optional
        Pixel distance of boundary. If no boundary width is given it is
        calculated as 2% of the image diagonal. The default is None.
    kernel_type : str, optional
        Type of kernel to use for boundary extraction. The default is 'box'.
    labels : torch.Tensor, optional
        Labels to include in IoU. The default is None.
    pos_label : int, optional
        Specifies in which dim labels are specified. The default is 1.
    batch_iou : bool, optional
        Calculate IoU for each sample. The default is False.
    non_presence_threshold : float, optional
        Threshold at which number of pixels to include. The default is None.
    reduction : str, optional
        Modality how to reduce result. The default is 'mean'.
    smooth : float, optional
        Soften the result. The default is 10e-6.

    Returns
    -------
    boundary_jaccard_index : torch.Tensor
        Boundary Intersection over Union. NaN if classes are not presence.

    References
    ----------
        .. [1] Bowen C. et al., "Boundary IoU: Improving Object-Centric Image
            Segmentation Evaluation", https://arxiv.org/abs/2103.16562

    """
    if boundary_width is None:
        boundary_width = torch.div(
            torch.min(
                torch.tensor(
                    target.shape[pos_label+1:],
                    requires_grad=False,
                    dtype=int,
                    device=target.device,
                )
            ),
            50,
            rounding_mode='floor',
        )

    if boundary_width == 0:
        raise ValueError(f"Boundary width {boundary_width} is not supported.")

    target = mask_to_boundary_region(
        mask=target,
        boundary_width=boundary_width,
        kernel_type=kernel_type,
        lower_threshold_num_elements=1,
    ) * target
    output = mask_to_boundary_region(
        mask=output,
        boundary_width=boundary_width,
        kernel_type=kernel_type,
        lower_threshold_num_elements=1,
    ) * output

    boundary_jaccard_index = intersection_over_union(
        output=output,
        target=target,
        labels=labels,
        pos_label=pos_label,
        batch_iou=batch_iou,
        non_presence_threshold=non_presence_threshold,
        reduction=reduction,
        smooth=smooth,
    )

    return boundary_jaccard_index
